/*
 * Copyright (c) 1996 Simon J. Gerraty, all rights reserved.
 * Copyright (c) 1983, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char RCSid[] = "$Id: sslrcmd.c,v 1.14 1997/05/17 12:55:17 sjg Exp $";
#endif

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/stat.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#include <signal.h>
#include <fcntl.h>
#include <netdb.h>
#include <unistd.h>
#include <pwd.h>
#include <errno.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#ifndef SA_NOCLDSTOP
# include "sigact.h"				/* <sjg> fake POSIX signals */
#endif

#ifdef sun
# include "sunos.h"
#endif

#include "sslfd.h"

#ifndef SSHELL_PORT
# define SSHELL_PORT	614			/* official port */
#endif

/*
 * <sjg> simplified rcmd()
 * We do not use suid, reserved ports etc,
 * since our cert proves who we are.
 */
int
ssl_rcmd(ahost, rport, locuser, remuser, cmd, fd2p)
	char **ahost;
	u_short rport;
	const char *locuser, *remuser, *cmd;
	int *fd2p;
{
	struct hostent *hp;
	struct sockaddr_in sin, from;
	fd_set reads;
	sigset_t mask, omask;
	pid_t pid;
	int s, lport = 0, timo;
	char c;
#ifdef ALT_SSHELL_PORT
	int altport = 0;
#endif
	
	pid = getpid();
	hp = gethostbyname(*ahost);
	if (hp == NULL) {
#if defined(BSD4_4) || defined(HAVE_HERROR)
		herror(*ahost);
#else
		fprintf(stderr, "ssl_rcmd: host '%s' unknown, or lookup failed.\n", *ahost);
#endif
		return (-1);
	}
	*ahost = hp->h_name;
	if (rport == 0 || rport == (u_short)-1) {
		struct servent *sp;
#ifdef ALT_SSHELL_PORT
		altport = htons(ALT_SSHELL_PORT);
#endif
		if ((sp = getservbyname("sshell", "tcp")) || /* official */
		    (sp = getservbyname("SSLshell", "tcp"))) { /* old name */
			rport = sp->s_port;
		} else {
			rport = htons(SSHELL_PORT);
		}
	}

	sigemptyset(&mask);
	sigaddset(&mask,SIGURG);
	(void) sigprocmask(SIG_BLOCK, &mask, &omask);

	for (timo = 1;;) {
		s = socket(AF_INET, SOCK_STREAM, 0);
		if (s < 0) {
			if (errno == EAGAIN)
				(void)fprintf(stderr,
				"ssl_rcmd: socket: All ports in use\n");
			else
				(void)fprintf(stderr, "ssl_rcmd: socket: %s\n",
					      strerror(errno));
			(void) sigprocmask(SIG_SETMASK, &omask, 0);
			return (-1);
		}
#ifdef F_SETOWN
		fcntl(s, F_SETOWN, pid);
#endif
#ifdef BSD4_4
		sin.sin_len = sizeof(struct sockaddr_in);
#endif
		sin.sin_family = hp->h_addrtype;
		sin.sin_port = rport;
		bcopy(hp->h_addr_list[0], &sin.sin_addr, hp->h_length);
		if (connect(s, (struct sockaddr *)&sin, sizeof(sin)) >= 0)
			break;
		(void)close(s);
		if (errno == ECONNREFUSED && timo <= 16) {
#ifdef ALT_SSHELL_PORT
			if (altport) {
				static int skip = 0;
				int oport;

				/*
				 * Ok, we want to toggle between the
				 * official and old port no.s, each
				 * time we use the old port no. we
				 * skip the sleep...
				 */
				oport = rport;
				rport = altport;
				altport = oport;
				if (!skip)	/* worth it? */
					skip = htons(ALT_SSHELL_PORT);
				if (rport == skip)
					continue; /* skip sleep */
			}
#endif
			(void)sleep(timo);
			timo *= 2;
			continue;
		}
		if (hp->h_addr_list[1] != NULL) {
			int oerrno = errno;

			(void)fprintf(stderr, "connect to address %s: ",
			    inet_ntoa(sin.sin_addr));
			errno = oerrno;
			perror(0);
			hp->h_addr_list++;
			bcopy(hp->h_addr_list[0], &sin.sin_addr, hp->h_length);
			(void)fprintf(stderr, "Trying %s...\n",
			    inet_ntoa(sin.sin_addr));
			continue;
		}
		(void)fprintf(stderr, "%s: %s\n", hp->h_name, strerror(errno));
		(void) sigprocmask(SIG_SETMASK, &omask, 0);
		return (-1);
	}

	if (ssl_start(s, -1, 0) < 0)
		goto bad;
	
	if (fd2p == 0) {
		write_net(s, "", 1);
		lport = 0;
	} else {
		/* <sjg> all quite different */
		char num[8];
		int nfd, s3, s2 = socket(AF_INET, SOCK_STREAM, 0);
		int len = sizeof(from);

		if (s2 < 0)
			goto bad;
		listen(s2, 1);
		bzero((char *)&from, sizeof(from));
		from.sin_family = AF_INET;
		from.sin_port = 0;
		if (getsockname(s2, (struct sockaddr *)&from, &len) < 0) {
			perror("getsockname(s2)");
			close(s2);
			goto bad;
		}
		lport =  ntohs(from.sin_port);
		(void)sprintf(num, "%d", lport);
		if (write_net(s, num, strlen(num)+1) != strlen(num)+1) {
			(void)fprintf(stderr,
			    "ssl_rcmd: write (setting up stderr): %s\n",
			    strerror(errno));
			(void)close(s2);
			goto bad;
		}
		FD_ZERO(&reads);
		FD_SET(s, &reads);
		FD_SET(s2, &reads);
		errno = 0;
		nfd = (s > s2) ? s : s2;
		if (select(nfd + 1, &reads, 0, 0, 0) < 1 ||
		    !FD_ISSET(s2, &reads)) {
			if (errno != 0)
				(void)fprintf(stderr,
				    "ssl_rcmd: select (setting up stderr): %s\n",
				    strerror(errno));
			else
				(void)fprintf(stderr,
				"select: protocol failure in circuit setup\n");
			(void)close(s2);
			goto bad;
		}
		len = sizeof(from);
		s3 = accept(s2, (struct sockaddr *)&from, &len);
		(void)close(s2);
		if (s3 < 0) {
			(void)fprintf(stderr,
			    "ssl_rcmd: accept: %s\n", strerror(errno));
			lport = 0;
			goto bad;
		}
		*fd2p = s3;
		if (ssl_clone(s3, s, -1, 0) < 0) /* uses previous setup */
			goto bad2;
	}
	
	(void)write_net(s, (char *) locuser, strlen(locuser)+1);
	(void)write_net(s, (char *) remuser, strlen(remuser)+1);
	(void)write_net(s, (char *) cmd, strlen(cmd)+1);

	if (read_net(s, &c, 1) != 1) {
		(void)fprintf(stderr,
		    "ssl_rcmd: %s: %s\n", *ahost, strerror(errno));
		goto bad2;
	}
	if (c != 0) {
		char buf[BUFSIZ];
		int i;
		
		if ((i = read_net(s, buf, sizeof(buf))) > 0) {
			(void)write(STDERR_FILENO, buf, i);
		}
		goto bad2;
	}
	(void) sigprocmask(SIG_SETMASK, &omask, 0);
	return (s);
bad2:
	if (lport)
		(void)ssl_close(*fd2p);
bad:
	(void)close(s);
	(void) sigprocmask(SIG_SETMASK, &omask, 0);
	return (-1);
}
